/*
 * livepatch_bsc1196301
 *
 * Fix for CVE-2022-25636, bsc#1196301
 *
 *  Upstream commit:
 *  b1a5983f56e3 ("netfilter: nf_tables_offload: incorrect flow offload action
 *                 array size")
 *
 *  SLE12-SP3 commit:
 *  not affected
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commit:
 *  not affected
 *
 *  SLE15-SP2 commit:
 *  not affected
 *
 *  SLE15-SP3 commit:
 *  30b89a9c7b1382ccaf93152b73afcae377454423
 *
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_NF_TABLES)

#if !IS_MODULE(CONFIG_NF_TABLES)
#error "Live patch supports only CONFIG_NF_TABLES=m"
#endif

#if !IS_MODULE(CONFIG_NFT_DUP_NETDEV)
#error "Live patch supports only CONFIG_NFT_DUP_NETDEV=m"
#endif

#if !IS_MODULE(CONFIG_NFT_FWD_NETDEV)
#error "Live patch supports only CONFIG_NFT_FWD_NETDEV=m"
#endif


/* klp-ccp: from net/netfilter/nft_immediate.c */
static const struct nft_expr_ops (*klpe_nft_imm_ops);

/* klp-ccp: from net/netfilter/nft_dup_netdev.c */
static const struct nft_expr_ops (*klpe_nft_dup_netdev_ops);

/* klp-ccp: from net/netfilter/nft_fwd_netdev.c */
static const struct nft_expr_ops (*klpe_nft_fwd_netdev_ops);

/* klp-ccp: from net/netfilter/nf_tables_offload.c */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <net/flow_offload.h>
#include <net/netfilter/nf_tables.h>
#include <net/netfilter/nf_tables_offload.h>
#include <net/netfilter/nf_tables_core.h>

/* New */
static bool klpp_nft_expr_ops_offload_action(const struct nft_expr *expr)
{
	const struct nft_expr_ops * const nft_imm_ops = READ_ONCE(klpe_nft_imm_ops);
	const struct nft_expr_ops * const nft_dup_netdev_ops =
		READ_ONCE(klpe_nft_dup_netdev_ops);
	const struct nft_expr_ops * const nft_fwd_netdev_ops =
		READ_ONCE(klpe_nft_fwd_netdev_ops);

	if (expr->ops == nft_dup_netdev_ops || expr->ops == nft_fwd_netdev_ops)
		return true;

	if (expr->ops == nft_imm_ops) {
		const struct nft_immediate_expr *priv = nft_expr_priv(expr);

		if(priv->dreg == NFT_REG_VERDICT)
			return true;

		return false;
	}

	return false;
}


/* klp-ccp: from include/net/netfilter/nf_tables_offload.h */
struct nft_flow_rule *klpp_nft_flow_rule_create(struct net *net, const struct nft_rule *rule);
static void (*klpe_nft_flow_rule_destroy)(struct nft_flow_rule *flow);

/* klp-ccp: from net/netfilter/nf_tables_offload.c */
static struct nft_flow_rule *nft_flow_rule_alloc(int num_actions)
{
	struct nft_flow_rule *flow;

	flow = kzalloc(sizeof(struct nft_flow_rule), GFP_KERNEL);
	if (!flow)
		return NULL;

	flow->rule = flow_rule_alloc(num_actions);
	if (!flow->rule) {
		kfree(flow);
		return NULL;
	}

	flow->rule->match.dissector	= &flow->match.dissector;
	flow->rule->match.mask		= &flow->match.mask;
	flow->rule->match.key		= &flow->match.key;

	return flow;
}

struct nft_flow_rule *klpp_nft_flow_rule_create(struct net *net,
					   const struct nft_rule *rule)
{
	struct nft_offload_ctx *ctx;
	struct nft_flow_rule *flow;
	int num_actions = 0, err;
	struct nft_expr *expr;

	expr = nft_expr_first(rule);
	while (expr->ops && expr != nft_expr_last(rule)) {
		/*
		 * Fix CVE-2022-25636
		 *  -1 line, +1 line
		 */
		if (klpp_nft_expr_ops_offload_action(expr))
			num_actions++;

		expr = nft_expr_next(expr);
	}

	if (num_actions == 0)
		return ERR_PTR(-EOPNOTSUPP);

	flow = nft_flow_rule_alloc(num_actions);
	if (!flow)
		return ERR_PTR(-ENOMEM);

	expr = nft_expr_first(rule);

	ctx = kzalloc(sizeof(struct nft_offload_ctx), GFP_KERNEL);
	if (!ctx) {
		err = -ENOMEM;
		goto err_out;
	}
	ctx->net = net;
	ctx->dep.type = NFT_OFFLOAD_DEP_UNSPEC;

	while (expr->ops && expr != nft_expr_last(rule)) {
		if (!expr->ops->offload) {
			err = -EOPNOTSUPP;
			goto err_out;
		}
		err = expr->ops->offload(ctx, flow, expr);
		if (err < 0)
			goto err_out;

		expr = nft_expr_next(expr);
	}
	flow->proto = ctx->dep.l3num;
	kfree(ctx);

	return flow;
err_out:
	kfree(ctx);
	(*klpe_nft_flow_rule_destroy)(flow);

	return ERR_PTR(err);
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1196301.h"
#include "../kallsyms_relocs.h"

#define LIVEPATCHED_MODULE "nf_tables"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "nft_flow_rule_destroy", (void *)&klpe_nft_flow_rule_destroy,
	  "nf_tables" },
	{ "nft_imm_ops", (void *)&klpe_nft_imm_ops, "nf_tables" },
};

static const struct nft_expr_ops (*__klpe_nft_dup_netdev_ops);

static struct klp_kallsyms_reloc klp_nft_dup_netdev_funcs[] = {
	{ "nft_dup_netdev_ops", (void *)&__klpe_nft_dup_netdev_ops,
	  "nft_dup_netdev" },
};

static const struct nft_expr_ops (*__klpe_nft_fwd_netdev_ops);

static struct klp_kallsyms_reloc klp_nft_fwd_netdev_funcs[] = {
	{ "nft_fwd_netdev_ops", (void *)&__klpe_nft_fwd_netdev_ops,
	  "nft_fwd_netdev" },
};

static int livepatch_bsc1196301_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret = 0;

	if (action == MODULE_STATE_COMING) {
		if (!strcmp(mod->name, LIVEPATCHED_MODULE)) {
			ret = __klp_resolve_kallsyms_relocs(klp_funcs,
							    ARRAY_SIZE(klp_funcs));

		} else if (!strcmp(mod->name, "nft_dup_netdev")) {
			ret = __klp_resolve_kallsyms_relocs(klp_nft_dup_netdev_funcs,
							    ARRAY_SIZE(klp_nft_dup_netdev_funcs));
			if (!ret)
				WRITE_ONCE(klpe_nft_dup_netdev_ops, __klpe_nft_dup_netdev_ops);

		} else if (!strcmp(mod->name, "nft_fwd_netdev")) {
			ret = __klp_resolve_kallsyms_relocs(klp_nft_fwd_netdev_funcs,
							    ARRAY_SIZE(klp_nft_fwd_netdev_funcs));
			if (!ret)
				WRITE_ONCE(klpe_nft_fwd_netdev_ops, __klpe_nft_fwd_netdev_ops);

		}

		WARN(ret, "livepatch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	} else if (action == MODULE_STATE_GOING) {
		if (!strcmp(mod->name, "nft_dup_netdev")) {
			__klpe_nft_dup_netdev_ops = NULL;
			WRITE_ONCE(klpe_nft_dup_netdev_ops, NULL);
		} else if (!strcmp(mod->name, "nft_fwd_netdev")) {
			__klpe_nft_fwd_netdev_ops = NULL;
			WRITE_ONCE(klpe_nft_fwd_netdev_ops, NULL);
		}
	}

	return ret;
}

static struct notifier_block livepatch_bsc1196301_module_nb = {
	.notifier_call = livepatch_bsc1196301_module_notify,
	.priority = INT_MIN+1,
};

int livepatch_bsc1196301_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(LIVEPATCHED_MODULE)) {
		ret = __klp_resolve_kallsyms_relocs(klp_funcs,
						    ARRAY_SIZE(klp_funcs));
		if (ret)
			goto out;
	}

	if (find_module("nft_dup_netdev")) {
		ret = __klp_resolve_kallsyms_relocs(klp_nft_dup_netdev_funcs,
						    ARRAY_SIZE(klp_nft_dup_netdev_funcs));
		if (ret)
			goto out;

		WRITE_ONCE(klpe_nft_dup_netdev_ops, __klpe_nft_dup_netdev_ops);
	}

	if (find_module("nft_fwd_netdev")) {
		ret = __klp_resolve_kallsyms_relocs(klp_nft_fwd_netdev_funcs,
						    ARRAY_SIZE(klp_nft_fwd_netdev_funcs));
		if (ret)
			goto out;

		WRITE_ONCE(klpe_nft_fwd_netdev_ops, __klpe_nft_fwd_netdev_ops);
	}

	ret = register_module_notifier(&livepatch_bsc1196301_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void livepatch_bsc1196301_cleanup(void)
{
	unregister_module_notifier(&livepatch_bsc1196301_module_nb);
}

#endif /* IS_ENABLED(CONFIG_NF_TABLES) */
